#ifndef sigma_Dag_INCLUDED
#define sigma_Dag_INCLUDED

#include <assert.h>

#include <deque>
#include <map>

namespace sigma {

template <typename Node, typename Edge, typename Compare=std::less<Node> >
struct Dag {

  // thrown if inserting edge will create circularity

  struct Circular {
    Circular(const Node& f, const Node& t) : from(f), to(t) {}
    Node from;
    Node to;
  }; 

  typedef std::deque<Edge> Path;
  typedef std::map<Node,Edge,Compare> EdgeMap;
  typedef std::map<Node,EdgeMap,Compare> Edges;

  inline void insert(const Node& from, 
		     const Node& to,
		     const Edge& edge) {
    assert(!(from==to));
    if (pathExists(to,from)) throw Circular(from, to);
    edges[from][to] = edge;
  }

  inline Path getPath(const Node& from, const Node& to) {
    assert(!(from==to));
    Path shortest_path;
    typename Path::size_type min_path_length = shortest_path.max_size();
    typename Edges::const_iterator i = edges.find(from);
    if (i==edges.end()) return shortest_path; // empty path
    for (typename EdgeMap::const_iterator j = (*i).second.begin();
	 j != (*i).second.end();
	 ++j) {

      if ((*j).first==to) {
	shortest_path.clear();
	shortest_path.push_back((*j).second);
	return shortest_path;
      }

      Path path = getPath((*j).first, to);
      typename Path::size_type path_length = path.size();
      if ((path.size()!=0) && (path_length<min_path_length)) {
	min_path_length = path_length;
	shortest_path = path;
	shortest_path.push_front((*j).second);
      }
    }
    return shortest_path;
  }

  inline bool pathExists(const Node& from, const Node& to) {
    assert(!(from==to));
    typename Edges::const_iterator i = edges.find(from);
    if (i==edges.end()) return false;
    for (typename EdgeMap::const_iterator j = (*i).second.begin();
	 j != (*i).second.end();
	 ++j) {

      if ((*j).first==to) return true;

      if (pathExists((*j).first, to)) return true;
    }
    return false;
  }

  Edges edges;
};

} // end namespace sigma

#endif
