#include "Generic.hpp"

namespace sigma {

Generic::Generic() 
  : branch(new generic_detail::GenericBranch()), 
    ref_count(new int(1)) {}
Generic::Generic(const std::string& n) 
  : name(n), 
    branch(new generic_detail::GenericBranch()), 
    ref_count(new int(1)) {}

Generic::Generic(const Generic& that) 
  : name(that.name),
    branch(that.branch), 
    ref_count(that.ref_count) {

  *ref_count = *ref_count + 1;
}


Generic& Generic::operator=(const Generic& that) {
  *(that.ref_count) = *(that.ref_count) + 1;
  *ref_count = *ref_count - 1;
  if (*ref_count == 0) {
    assert(branch);
    branch->cleanup();
    delete branch;
  }
  ref_count = that.ref_count;
  branch = that.branch;
  name = that.name;

  return *this;
}
  

Generic::~Generic() {

  *ref_count = *ref_count - 1;
  if (*ref_count == 0) {
    assert(branch);
    branch->cleanup();
    delete branch;
  }
}


Generic copy(const Generic& that) {

  Generic generic(that.name);
  assert(that.branch);
  generic.branch = that.branch->copy();
  return generic;
}


Forms Generic::getForms() const {

  Forms forms;
  assert(branch);
  branch->getForms(forms);
  return forms;
}

namespace generic_detail {

void GenericBranch::cleanup() {

  for (GenericBranches::iterator i = branches.begin();
       i != branches.end();
       ++i) {
    GenericBranch* branch = (*i).second;
    branch->cleanup();
    delete branch;
  }
}

GenericBranch* GenericBranch::copy() const {

  GenericBranch* branch = new GenericBranch();
  branch->method = method;
  for (GenericBranches::const_iterator i = branches.begin();
       i != branches.end();
       ++i) {
    std::pair<Type,GenericBranch*> p((*i).first,(*i).second->copy());
    branch->branches.push_back(p);
  }
  return branch;
}
  

Ptr<Method> GenericBranch::dispatch(Args::const_iterator i, 
				    Args::const_iterator end) const {

  if (i==end) return method;

  for (GenericBranches::const_iterator j = branches.begin();
       j != branches.end();
       ++j) {

    if (isa((*i).type(), (*j).first)) {
      GenericBranch* next_branch = (*j).second;
      Args::const_iterator tmp = i;
      ++tmp;
      Ptr<Method> f = next_branch->dispatch(tmp, end);
      if (!null(f)) return f;
    }
  }

  return Ptr<Method>();
}


void GenericBranch::bind(const Ptr<Method>& f, unsigned int i) {

  if (i==f->arity()) {
    method = f;
    return;
  }

  Type type = f->argType(i);

  GenericBranches::iterator j = branches.begin();

  while(j!=branches.end() && !isa(type,(*j).first)) ++j;

  if ((j==branches.end()) || 
      (!(type==(*j).first))) {
    GenericBranch* branch = new GenericBranch();
    j = branches.insert(j, std::make_pair(Type(type), branch));
  }
  GenericBranch* next_branch = (*j).second;
  next_branch->bind(f, ++i);
}


void GenericBranch::getForms(Forms& forms) const {

  if (!null(method)) {
    forms.push_back(Form(method->returnType(),
			 method->getArgTypes()));
  }
  for (GenericBranches::const_iterator i = branches.begin();
       i != branches.end();
       ++i) {

    (*i).second->getForms(forms);
  }
}


bool GenericBranch::contains(Cell const * const that_cell) const {

  if (!null(method) && method.package->contains(that_cell)) return true;

  for (GenericBranches::const_iterator i = branches.begin();
       i != branches.end();
       ++i) {

    GenericBranch* branch = (*i).second;
    if (branch->contains(that_cell)) return true;
  }
  return false;
}

} // end namespace generic_detail

} // end namespace sigma
