#ifndef sigma_Generic_INCLUDED
#define sigma_Generic_INCLUDED

#include <string>
#include <list>

#include <ptr/Type.hpp>
#include <ptr/Ptr.hpp>
#include <generic/Function.hpp>
#include <generic/Method.hpp>

namespace sigma {

namespace generic_detail {

struct GenericBranch;

typedef std::list<std::pair<Type,GenericBranch*> > GenericBranches;

struct GenericBranch {

  void cleanup();
  Ptr<Method> dispatch(Args::const_iterator i, 
			 Args::const_iterator end) const;
  void bind(const Ptr<Method>&, unsigned int i);

  bool contains(Cell const * const) const;

  GenericBranch* copy() const;

  void getForms(Forms& forms) const;

  Ptr<Method> method;
  GenericBranches branches;
};

} // end namespace generic_detail

struct Generic : Function {

  struct NoMatch {
    NoMatch(const std::string& n,
	    const Types& t,
	    const Forms& f) : name(n), types(t), forms(f) {}
    std::string name;
    Types types;
    Forms forms;
  };

  Generic();
  Generic(const std::string& name);	// for annotation
  Generic(const Generic&);
  Generic& operator=(const Generic&);
  virtual ~Generic();

  friend Generic copy(const Generic&);

  inline std::string getName() const {return name;}
  Forms getForms() const;

  inline Ptr<Method> dispatch(const Args& args) const {

    if (branch) 
      return branch->dispatch(args.begin(), args.end());
    return Ptr<Method>();
  }


  inline void bind(const Ptr<Method>& f) {
    if (null(f)) return;
    if (!branch) branch = new generic_detail::GenericBranch();
    branch->bind(f, 0);
  }

  inline Ptr<Anything> apply(const Args& args) const {
    Ptr<Method> f = dispatch(args);
    if (null(f)) throw NoMatch(name, getTypes(args), getForms());
    return f->apply(args);
  }

  inline Ptr<Anything> operator()() const {
    Args args;
    return apply(args);
  }

  template <typename A1>
  inline Ptr<Anything> operator()(const Ptr<A1>& a1) const {
    Args args;
    args.push_back(a1);
    return apply(args);
  }

  template <typename A1, typename A2>
  inline Ptr<Anything> operator() (const Ptr<A1>& a1, 
			  const Ptr<A2>& a2) const {
    Args args;
    args.push_back(a1);
    args.push_back(a2);
    return apply(args);
  }

  template <typename A1, typename A2, typename A3>
  inline Ptr<Anything> operator() (const Ptr<A1>& a1, 
				   const Ptr<A2>& a2,
				   const Ptr<A3>& a3) const {
    Args args;
    args.push_back(a1);
    args.push_back(a2);
    args.push_back(a3);
    return apply(args);
  }

  inline bool contains(Cell const * const that_cell) {
    return (branch && branch->contains(that_cell));
  }

 private:

  std::string name;
  generic_detail::GenericBranch* branch;
  int* ref_count;
};

namespace cell_detail {

template<>
struct TypedPackage<Generic> : TypedPackageBase<Generic> {

  explicit TypedPackage(Generic* ptr) :
    TypedPackageBase<Generic>(ptr) {}

  virtual bool contains(Cell const * const that_cell) const {
    Generic* generic = reinterpret_cast<Generic*>(ptr);
    return generic->contains(that_cell);
  }
};

} // end namespace cell

namespace {
DeclareTypeName<Generic> generic_type_name("generic");
DeclareSubtypeConverter<Generic,Function> generic_function_type_converter;
}


} // end namespace sigma

#endif
