#ifndef sigma_Primitive_INCLUDED
#define sigma_Primitive_INCLUDED

#include <typeinfo>

#include <ptr/Ptr.hpp>
#include <generic/Method.hpp>

namespace sigma {

template <typename FuncPtr> struct Primitive;

template <typename R>
struct Primitive<Ptr<R>(*)()> : Method {

  typedef Ptr<R>(*FuncPtr)();

  Primitive(FuncPtr f) : func_ptr(f) {}

  enum {ARITY=0};
  virtual unsigned int arity() const {return ARITY;}
  virtual Type returnType() const {return Type(typeid(R));}
  virtual Type argType(unsigned int) const {return Type(typeid(Nothing));}
  virtual Ptr<Anything> apply(const Args& args) const {
    assert(args.size()==ARITY);
    return (*func_ptr)();
  }

  FuncPtr func_ptr;
};


template <typename R, typename A0>
struct Primitive<Ptr<R>(*)(const Ptr<A0>&)> : Method {

  typedef Ptr<R>(*FuncPtr)(const Ptr<A0>&);

  Primitive(FuncPtr f) : func_ptr(f) {}

  enum {ARITY=1};
  virtual unsigned int arity() const {return ARITY;}
  virtual Type returnType() const {return Type(typeid(R));}
  virtual Type argType(unsigned int i) const {
    switch(i) {
    case 0:
      return Type(typeid(A0));
    default:
      return Type(typeid(Nothing));
    }
  }
  virtual Ptr<Anything> apply(const Args& args) const {
    assert(args.size()==ARITY);
    return (*func_ptr)(args[0]);
  }
  
  FuncPtr func_ptr;
};


template <typename R, typename A0, typename A1>
struct Primitive<Ptr<R>(*)(const Ptr<A0>&,
			   const Ptr<A1>&)> : Method {

  typedef Ptr<R>(*FuncPtr)(const Ptr<A0>&,
			   const Ptr<A1>&);

  Primitive(FuncPtr f) : func_ptr(f) {}

  enum {ARITY=2};
  virtual unsigned int arity() const {return ARITY;}
  virtual Type returnType() const {return Type(typeid(R));}
  virtual Type argType(unsigned int i) const {
    switch(i) {
    case 0:
      return Type(typeid(A0));
    case 1:
      return Type(typeid(A1));
    default:
      return Type(typeid(Nothing));
    }
  }
  virtual Ptr<Anything> apply(const Args& args) const {
    assert(args.size()==ARITY);
    return (*func_ptr)(args[0], args[1]);
  }
  
  FuncPtr func_ptr;
};


template <typename R, typename A0, typename A1, typename A2>
struct Primitive<Ptr<R>(*)(const Ptr<A0>&,
			   const Ptr<A1>&,
			   const Ptr<A2>&)> : Method {

  typedef Ptr<R>(*FuncPtr)(const Ptr<A0>&,
			   const Ptr<A1>&,
			   const Ptr<A2>&);

  Primitive(FuncPtr f) : func_ptr(f) {}

  enum {ARITY=3};
  virtual unsigned int arity() const {return ARITY;}
  virtual Type returnType() const {return Type(typeid(R));}
  virtual Type argType(unsigned int i) const {
    switch(i) {
    case 0:
      return Type(typeid(A0));
    case 1:
      return Type(typeid(A1));
    case 2:
      return Type(typeid(A2));
    default:
      return Type(typeid(Nothing));
    }
  }
  virtual Ptr<Anything> apply(const Args& args) const {
    assert(args.size()==ARITY);
    return (*func_ptr)(args[0], args[1], args[2]);
  }
  
  FuncPtr func_ptr;
};


template <typename R>
Ptr<Primitive<Ptr<R>(*)()> > primitive(Ptr<R>(*f)()) {
  return new Primitive<Ptr<R>(*)()>(f);
}

template <typename R, typename A0>
Ptr<Primitive<Ptr<R>(*)(const Ptr<A0>&)> > primitive(Ptr<R>(*f)(const Ptr<A0>&)) {
  return new Primitive<Ptr<R>(*)(const Ptr<A0>&)>(f);
}

template <typename R, typename A0, typename A1>
Ptr<Primitive<Ptr<R>(*)(const Ptr<A0>&, 
			const Ptr<A1>&)> > primitive(Ptr<R>(*f)(const Ptr<A0>&,
								const Ptr<A1>&)) {
  return new Primitive<Ptr<R>(*)(const Ptr<A0>&,
				 const Ptr<A1>&)>(f);
}


template <typename R, typename A0, typename A1, typename A2>
Ptr<Primitive<Ptr<R>(*)(const Ptr<A0>&, 
			const Ptr<A1>&,
			const Ptr<A2>&)> > primitive(Ptr<R>(*f)(const Ptr<A0>&,
								const Ptr<A1>&,
								const Ptr<A2>&)) {
  return new Primitive<Ptr<R>(*)(const Ptr<A0>&,
				 const Ptr<A1>&,
				 const Ptr<A2>&)>(f);
}

} // end namespace sigma

#endif
