#ifndef sigma_lex_Dstate_INCLUDED
#define sigma_lex_Dstate_INCLUDED


#include <map>

#include <lex/Pos.hpp>

namespace sigma {

namespace dfa_detail {

template <typename Letter>
struct PosSets : std::map<PosSet<Letter>,size_t> {};

template <typename Letter>
struct Dstate : PosSets<Letter>::const_iterator {

  Dstate() {} // needed for operator[] use by Dtrans
  Dstate(typename PosSets<Letter>::const_iterator that) 
    : PosSets<Letter>::const_iterator(that) {}
};


template <typename Letter>
bool operator<(Dstate<Letter> a, Dstate<Letter> b) {
  return ((*a).second < (*b).second);
}

template <typename Letter>
struct MakeDstate {

  /*
    The bool indicates whether the Dstate returned is new. The make
    dfa function uses this to determine whether to put the dstate into
    the unmarked set.
  */

  std::pair<Dstate<Letter>,bool> operator()(const PosSet<Letter>& ps) {

    typename PosSets<Letter>::const_iterator i = pos_sets.find(ps);
    if (i != pos_sets.end()) {
      return std::pair<Dstate<Letter>,bool>(i,false);
    }

    return pos_sets.insert(std::make_pair(ps,pos_sets.size()));
  }

  PosSets<Letter> pos_sets;

  const PosSets<Letter>& getPosSets() const {return pos_sets;}
};

} // end namespace dfa_detail

} // end namespace sigma

#endif
