#ifndef sigma_lex_Lexer_INCLUDED
#define sigma_lex_Lexer_INCLUDED

#include <deque>

#include <lex/Dfa.hpp>

namespace sigma {

template <typename Letter, 
	  typename Token, 
	  typename Context, 
	  typename InputIterator>
struct Lexer {

  Lexer(Ptr<Dfa<Letter,Token,Context> > const & d,
	Ptr<Context> const & c,
	InputIterator& i, 
	InputIterator& e)
    : dfa(d), context(c), in(i), end(e) {
  }

  Ptr<Dfa<Letter,Token,Context> > dfa;
  Ptr<Context> context;
  InputIterator& in;
  InputIterator& end;

  std::deque<Token> peek_stack;

  Token lex() {

    if (peek_stack.empty()) return dfa->lex(context,in,end);
    Token result = peek_stack.front();
    peek_stack.pop_front();
    return result;
  }

  Token peek(unsigned int n=0) {
    for (unsigned int i=peek_stack.size(); 
	 i<(n+1);
	 ++i) {

      peek_stack.push_back(dfa->lex(context,in,end));
    }
    assert(n<peek_stack.size());
    return peek_stack[n];
  }

  void pop(unsigned int n=1) {
    assert(n<=peek_stack.size());
    for (unsigned int i = 0; i<n; ++i) peek_stack.pop_front();
  }

};


} // namespace sigma

#endif
