#ifndef sigma_Rational_INCLUDED
#define sigma_Rational_INCLUDED

#include <iosfwd>
#include <number/Sign.hpp>
#include <number/Natural.hpp>
#include <number/Integer.hpp>

namespace sigma {

template <Base base=256>
struct Rational {

  Integer<base> num;
  Natural<base> den;

  Rational(const Rational<base>& that)
    : num(that.num), den(that.den) {
    reduce();	// is this necessary?
  }

  template <Base that_base>
  Rational(const Rational<that_base>& that)
    : num(that.num), den(that.den) {
    reduce();	// is this necessary?
  }


  Rational(const Integer<base>& n, const Natural<base>& d) : num(n), den(d) {
    reduce();
  }

  Rational(const Integer<base>& n, const Integer<base>& d) 
    : num(n.natural, n.sign * d.sign), den(d.natural) {
    reduce();
  }

  Rational(const Integer<base>& n) : num(n), den(1) {}
  Rational(const Natural<base>& n) : num(n), den(1) {}
  explicit Rational(const Digits& digits) : num(digits), den(1) {}

  Rational(long n) : num(n), den(1) {}
  Rational(long long n) : num(n), den(1) {}

  Rational() : num(0), den(1) {}

  void reduce() {
    if (!den.isZero()) {
      Natural<base> d = gcd(num.natural, den);
      num = num / Integer<base>(d);
      den = den / d;
    }
  }

  operator double() const {
    // Good enough for now.
    return double(num)/double(den);
  }
};


template <Base base>
Rational<base> operator+(const Rational<base>& a, const Rational<base>& b) {
  Rational<base> result(a.num * b.den + b.num * a.den,
			a.den * b.den);
  return result;
}

template <Base base>
Rational<base> operator-(const Rational<base>& a, const Rational<base>& b) {
  Rational<base> result(a.num * b.den - b.num * a.den,
			a.den * b.den);
  return result;
}

template <Base base>
Rational<base> operator*(const Rational<base>& a, const Rational<base>& b) {
  Rational<base> result(a.num * b.num, a.den * b.den);
  return result;
}

template <Base base>
Rational<base> operator/(const Rational<base>& a, const Rational<base>& b) {
  Rational<base> result(a.num * b.den, b.num * a.den);
  return result;
}

template <Base base>
bool operator<(const Rational<base>& a, const Rational<base>& b) {
  return (a.num * b.den) < (b.num * a.den);
}

template <Base base>
bool operator==(const Rational<base>& a, const Rational<base>& b) {
  // assumes always in reduced form
  return (a.num == b.num) && (a.den == b.den);
}



template <Base base>
Rational<base> operator-(const Rational<base>& that) {
  Rational<base> result(-that.num, that.den);
  return result;
}


std::ostream& operator<<(std::ostream&, const Rational<10>&);
std::ostream& operator<<(std::ostream&, const Rational<256>&);

} // end namespace sigma

#endif
