#ifndef sigma_parse_Action_INCLUDED
#define sigma_parse_Action_INCLUDED

#include <iostream>

namespace sigma {

namespace parse_detail {

struct Shift {
  Shift(int s) : shift(s) {}
  int shift;
};

struct Reduce {
  Reduce(int r) : reduce(r) {}
  int reduce;
};

struct GoTo {
  GoTo(int s) : go_to(s) {}
  int go_to;
};

struct Accept {};

struct Action {

  Action(const Action& that) 
    : type(that.type), state_or_rule(that.state_or_rule) {}
  Action(Shift s) : type(SHIFT), state_or_rule(s.shift) {}
  Action(Reduce r) : type(REDUCE), state_or_rule(r.reduce) {}
  Action(GoTo s) : type(GOTO), state_or_rule(s.go_to) {}
  Action(Accept) : type(ACCEPT), state_or_rule(-1) {}

  enum ActionType {SHIFT, REDUCE, GOTO, ACCEPT} type;
  int state_or_rule;

};


inline bool operator==(const Action& a, const Action& b) {
  return ((a.state_or_rule==a.state_or_rule) &&
	  (a.type==b.type));
}

inline bool operator!=(const Action& a, const Action& b) {
  return !(a==b);
}


inline std::ostream& operator<<(std::ostream& out, const Action& action) {

  switch (action.type) {
  case Action::SHIFT:
    return out << "shift " << action.state_or_rule;
  case Action::REDUCE:
    return out << "reduce " << action.state_or_rule;
  case Action::GOTO:
    return out << "goto " << action.state_or_rule;
  case Action::ACCEPT:
    return out << "accept";
  }

  return out << "unknown action";
}

} // namespace parse_detail

} // namespace sigma

#endif
