#ifndef sigma_parse_Item_INCLUDED
#define sigma_parse_Item_INCLUDED

#include <iostream>
#include <set>

#include <parse/Production.hpp>

namespace sigma {

namespace parse_detail {

template <typename Context>
struct Item {

  typedef typename std::set<Item<Context> > ItemSet;

  Item(int d, const Ptr<Production<Context> >& p) : dot(d), production(p) {}
  Item(const Item& that) : dot(that.dot), production(that.production) {}

  Symbol pending() const {
    if (dot<production->getArity()) {
      return production->argName(dot);
    }
    return Symbol();
  }

  Item advance() const {
    assert(dot<production->getArity());
    return Item(dot+1, production);
  }

  unsigned int dot;
  Ptr<Production<Context> > production;
};


template <typename Context>
bool operator<(const Item<Context>& a, const Item<Context>& b) {

  if (a.dot < b.dot) return true;
  if (a.dot > b.dot) return false;
  return (*a.production) < (*b.production);
}

template <typename Context>
std::ostream& operator<<(std::ostream& out, const Item<Context>& item) {

  out << item.production->returnName() << " -> ";
  for (unsigned int i=0; i<item.production->getArity(); ++i) {
    if (item.dot==i) out << ". ";
    out << item.production->argName(i) << " ";
  }
  if (item.dot==item.production->getArity()) out << ".";
  return out;
}

template <typename Context>
std::ostream& operator<<(std::ostream& out, 
			 const std::set<Item<Context> >& item_set) {

  typedef std::set<Item<Context> > ItemSet;

  for (typename ItemSet::const_iterator i = item_set.begin();
       i != item_set.end();
       ++i) {

    out << (*i) << std::endl;
  }
  return out;
}


} // end namespace parse_detail

} // end namespace sigma

#endif
