#ifndef sigma_parse_ProdTraits_INCLUDED
#define sigma_parse_ProdTraits_INCLUDED

#include <generic/Args.hpp>

namespace sigma {


template <typename Context, int>
struct ProdTraits;

template <typename Context>
struct ProdTraits<Context, 0> {
  typedef Cell (*FuncPtr)(const Ptr<Context>&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==0);
    return (*function)(context);
  }
};

template <typename Context>
struct ProdTraits<Context, 1> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==1);
    return (*function)(context,
		       args[0]);
  }
};

template <typename Context>
struct ProdTraits<Context, 2> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&, 
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==2);
    return (*function)(context,
		       args[0], 
		       args[1]);
  }
};


template <typename Context>
struct ProdTraits<Context, 3> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&, 
			 const Cell&, 
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==3);
    return (*function)(context,
		       args[0], 
		       args[1],	
		       args[2]);
  }
};


template <typename Context>
struct ProdTraits<Context, 4> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==4);
    return (*function)(context,
		       args[0], 
		       args[1],	
		       args[2],	
		       args[3]);
  }
};


template <typename Context>
struct ProdTraits<Context, 5> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==5);
    return (*function)(context,
		       args[0], 
		       args[1],	
		       args[2],	
		       args[3],	
		       args[4]);
  }
};


template <typename Context>
struct ProdTraits<Context, 6> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==6);
    return (*function)(context,
		       args[0], 
		       args[1],	
		       args[2],	
		       args[3],	
		       args[4],	
		       args[5]);
  }
};


template <typename Context>
struct ProdTraits<Context, 7> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==7);
    return (*function)(context,
		       args[0], 
		       args[1],	
		       args[2],	
		       args[3],	
		       args[4],	
		       args[5],	
		       args[6]);
  }
};

template <typename Context>
struct ProdTraits<Context, 8> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&, 
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==8);
    return (*function)(context,
		       args[0], 
		       args[1],	
		       args[2],	
		       args[3],	
		       args[4],	
		       args[5],	
		       args[6],	
		       args[7]);
  }
};


template <typename Context>
struct ProdTraits<Context, 9> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==9);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8]);
  }
};


template <typename Context>
struct ProdTraits<Context, 10> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==10);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9]);
  }
};


template <typename Context>
struct ProdTraits<Context, 11> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==11);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10]);
  }
};


template <typename Context>
struct ProdTraits<Context, 12> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==12);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10],
		       args[11]);
  }
};


template <typename Context>
struct ProdTraits<Context, 13> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==13);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10],
		       args[11],
		       args[12]);
  }
};


template <typename Context>
struct ProdTraits<Context, 14> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==14);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10],
		       args[11],
		       args[12],
		       args[13]);
  }
};


template <typename Context>
struct ProdTraits<Context, 15> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==15);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10],
		       args[11],
		       args[12],
		       args[13],
		       args[14]);
  }
};


template <typename Context>
struct ProdTraits<Context, 16> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==16);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10],
		       args[11],
		       args[12],
		       args[13],
		       args[14],
		       args[15]);
  }
};


template <typename Context>
struct ProdTraits<Context, 17> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==17);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10],
		       args[11],
		       args[12],
		       args[13],
		       args[14],
		       args[15],
		       args[16]);
  }
};


template <typename Context>
struct ProdTraits<Context, 18> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==18);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10],
		       args[11],
		       args[12],
		       args[13],
		       args[14],
		       args[15],
		       args[16],
		       args[17]);
  }
};


template <typename Context>
struct ProdTraits<Context, 19> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==19);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10],
		       args[11],
		       args[12],
		       args[13],
		       args[14],
		       args[15],
		       args[16],
		       args[17],
		       args[18]);
  }
};


template <typename Context>
struct ProdTraits<Context, 20> {
  typedef Cell(*FuncPtr)(const Ptr<Context>&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&,
			 const Cell&);
  static Cell apply(FuncPtr function, 
		    const Ptr<Context>& context, 
		    const Args& args) {
    assert(args.size()==20);
    return (*function)(context,
		       args[0],
		       args[1],
		       args[2],
		       args[3],
		       args[4],
		       args[5],
		       args[6],
		       args[7],
		       args[8],
		       args[9],
		       args[10],
		       args[11],
		       args[12],
		       args[13],
		       args[14],
		       args[15],
		       args[16],
		       args[17],
		       args[18],
		       args[19]);
  }
};

} // end namespace sigma


#endif
