#ifndef sigma_parse_Production_INCLUDED
#define sigma_parse_Production_INCLUDED

#include <ptr/Cell.hpp>
#include <ptr/Ptr.hpp>

#include <symbol/Symbol.hpp>
#include <generic/Args.hpp>
#include <parse/ProdTraits.hpp>

namespace sigma {

template <typename Context>
struct Production {

  Production(Symbol n, int a) : arity(a), return_name(n) {}
  inline Symbol returnName() const {return return_name;}
  inline unsigned int getArity() const {return arity;}

  // for lexographical compare

  virtual const Symbol* begin() const = 0;
  virtual const Symbol* end() const = 0;

  virtual Symbol argName(int index) const = 0;
  virtual Cell apply(const Ptr<Context>& context, const Args&) const = 0;

 private:

  unsigned int arity;
  Symbol return_name;
};

template <typename Context>
bool operator<(const Production<Context>& a, 
	       const Production<Context>& b) {

  if (a.returnName()<b.returnName()) return true;
  if (b.returnName()<a.returnName()) return false;
  return std::lexicographical_compare(a.begin(), a.end(),
				      b.begin(), b.end());
}

template <typename Context, int ARITY>
struct ProdImp : Production<Context> {

  typedef typename ProdTraits<Context, ARITY>::FuncPtr FuncPtr;

  ProdImp(Symbol n, FuncPtr f) 
    : Production<Context>(n,ARITY), function(f) {}

  virtual const Symbol* begin() const {
    return &arg_names[0];
  }

  virtual const Symbol* end() const {
    return &arg_names[ARITY];
  }

  virtual Symbol argName(int index) const {
    assert((index>=0) && (index<ARITY));
    return arg_names[index];
  }

  virtual Cell apply(const Ptr<Context>& context, const Args& args) const {
    assert(args.size()==ARITY);
    return ProdTraits<Context, ARITY>::apply(function, context, args);
  }

  // Arrays of zero elements are not allowed. This serendipitously
  // helps guarantee that no productions are nullable.

  Symbol arg_names[ARITY];
  FuncPtr function;
};


} // end namespace sigma


#endif
