#ifndef sigma_parse_Productions_INCLUDED
#define sigma_parse_Productions_INCLUDED

#include <set>
#include <map>

#include <parse/Production.hpp>
#include <parse/Item.hpp>
#include <parse/Edge.hpp>

namespace sigma {

namespace parse_detail {

typedef std::set<Symbol> SymbolSet;
typedef std::map<Symbol, SymbolSet> SymbolSetMap;


// returns whether or not set a changed

template <typename T> 
bool addTo(std::set<T>& a, const std::set<T>& b) {
  int old_size = a.size();
  a.insert(b.begin(), b.end());
  int new_size = a.size();
  return (new_size!=old_size);
}

template <typename Context>
static Cell pa_start(const Ptr<Context>& context, 
		     const Cell& a, 
		     const Cell&) {
  return a;
}


} // namespace parse_detail


template <typename Context>
struct Productions {

  typedef parse_detail::Item<Context> Item;
  typedef typename std::set<Item> ItemSet;
  typedef parse_detail::SymbolSet SymbolSet;
  typedef parse_detail::SymbolSetMap SymbolSetMap;

  typedef std::multimap<Symbol,Ptr<Production<Context> > > ProdMap;
  typedef typename ProdMap::const_iterator Iterator;
  typedef std::pair<Iterator,Iterator> Range;
  typedef std::pair<Symbol,Ptr<Production<Context> > > Entry;

  Productions(Symbol s, Symbol e)
    : start_symbol(s),
      end_symbol(e),
      start_prime(gensym("start")) {

    add(start_prime, &parse_detail::pa_start, start_symbol, end_symbol);
  }

  ProdMap prod_map;

  inline void add(Symbol result, typename ProdImp<Context,1>::FuncPtr f,
		  Symbol a0) {
    ProdImp<Context,1>* prodimp = new ProdImp<Context,1>(result,f);
    prodimp->arg_names[0] = a0;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,2>::FuncPtr f,
		  Symbol a0,
		  Symbol a1) {
    ProdImp<Context,2>* prodimp = new ProdImp<Context,2>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,3>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2) {
    ProdImp<Context,3>* prodimp = new ProdImp<Context,3>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,4>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3) {
    ProdImp<Context,4>* prodimp = new ProdImp<Context,4>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,5>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4) {
    ProdImp<Context,5>* prodimp = new ProdImp<Context,5>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,6>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5) {
    ProdImp<Context,6>* prodimp = new ProdImp<Context,6>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,7>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6) {
    ProdImp<Context,7>* prodimp = new ProdImp<Context,7>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prod_map.insert(Entry(result, prodimp));
  }


  inline void add(Symbol result, typename ProdImp<Context,8>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7) {
    ProdImp<Context,8>* prodimp = new ProdImp<Context,8>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,9>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8) {
    ProdImp<Context,9>* prodimp = new ProdImp<Context,9>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,10>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9) {
    ProdImp<Context,10>* prodimp = new ProdImp<Context,10>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,11>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10) {
    ProdImp<Context,11>* prodimp = new ProdImp<Context,11>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prod_map.insert(Entry(result, prodimp));
  }


  inline void add(Symbol result, typename ProdImp<Context,12>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10,
		  Symbol a11) {
    ProdImp<Context,12>* prodimp = new ProdImp<Context,12>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prodimp->arg_names[11] = a11;
    prod_map.insert(Entry(result, prodimp));
  }


  inline void add(Symbol result, typename ProdImp<Context,13>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10,
		  Symbol a11,
		  Symbol a12) {
    ProdImp<Context,13>* prodimp = new ProdImp<Context,13>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prodimp->arg_names[11] = a11;
    prodimp->arg_names[12] = a12;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,14>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10,
		  Symbol a11,
		  Symbol a12,
		  Symbol a13) {
    ProdImp<Context,14>* prodimp = new ProdImp<Context,14>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prodimp->arg_names[11] = a11;
    prodimp->arg_names[12] = a12;
    prodimp->arg_names[13] = a13;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,15>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10,
		  Symbol a11,
		  Symbol a12,
		  Symbol a13,
		  Symbol a14) {
    ProdImp<Context,15>* prodimp = new ProdImp<Context,15>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prodimp->arg_names[11] = a11;
    prodimp->arg_names[12] = a12;
    prodimp->arg_names[13] = a13;
    prodimp->arg_names[14] = a14;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,16>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10,
		  Symbol a11,
		  Symbol a12,
		  Symbol a13,
		  Symbol a14,
		  Symbol a15) {
    ProdImp<Context,16>* prodimp = new ProdImp<Context,16>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prodimp->arg_names[11] = a11;
    prodimp->arg_names[12] = a12;
    prodimp->arg_names[13] = a13;
    prodimp->arg_names[14] = a14;
    prodimp->arg_names[15] = a15;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,17>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10,
		  Symbol a11,
		  Symbol a12,
		  Symbol a13,
		  Symbol a14,
		  Symbol a15,
		  Symbol a16) {
    ProdImp<Context,17>* prodimp = new ProdImp<Context,17>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prodimp->arg_names[11] = a11;
    prodimp->arg_names[12] = a12;
    prodimp->arg_names[13] = a13;
    prodimp->arg_names[14] = a14;
    prodimp->arg_names[15] = a15;
    prodimp->arg_names[16] = a16;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,18>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10,
		  Symbol a11,
		  Symbol a12,
		  Symbol a13,
		  Symbol a14,
		  Symbol a15,
		  Symbol a16,
		  Symbol a17) {
    ProdImp<Context,18>* prodimp = new ProdImp<Context,18>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prodimp->arg_names[11] = a11;
    prodimp->arg_names[12] = a12;
    prodimp->arg_names[13] = a13;
    prodimp->arg_names[14] = a14;
    prodimp->arg_names[15] = a15;
    prodimp->arg_names[16] = a16;
    prodimp->arg_names[17] = a17;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,19>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10,
		  Symbol a11,
		  Symbol a12,
		  Symbol a13,
		  Symbol a14,
		  Symbol a15,
		  Symbol a16,
		  Symbol a17,
		  Symbol a18) {
    ProdImp<Context,19>* prodimp = new ProdImp<Context,19>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prodimp->arg_names[11] = a11;
    prodimp->arg_names[12] = a12;
    prodimp->arg_names[13] = a13;
    prodimp->arg_names[14] = a14;
    prodimp->arg_names[15] = a15;
    prodimp->arg_names[16] = a16;
    prodimp->arg_names[17] = a17;
    prodimp->arg_names[18] = a18;
    prod_map.insert(Entry(result, prodimp));
  }

  inline void add(Symbol result, typename ProdImp<Context,20>::FuncPtr f,
		  Symbol a0,
		  Symbol a1,
		  Symbol a2,
		  Symbol a3,
		  Symbol a4,
		  Symbol a5,
		  Symbol a6,
		  Symbol a7,
		  Symbol a8,
		  Symbol a9,
		  Symbol a10,
		  Symbol a11,
		  Symbol a12,
		  Symbol a13,
		  Symbol a14,
		  Symbol a15,
		  Symbol a16,
		  Symbol a17,
		  Symbol a18,
		  Symbol a19) {
    ProdImp<Context,20>* prodimp = new ProdImp<Context,20>(result,f);
    prodimp->arg_names[0] = a0;
    prodimp->arg_names[1] = a1;
    prodimp->arg_names[2] = a2;
    prodimp->arg_names[3] = a3;
    prodimp->arg_names[4] = a4;
    prodimp->arg_names[5] = a5;
    prodimp->arg_names[6] = a6;
    prodimp->arg_names[7] = a7;
    prodimp->arg_names[8] = a8;
    prodimp->arg_names[9] = a9;
    prodimp->arg_names[10] = a10;
    prodimp->arg_names[11] = a11;
    prodimp->arg_names[12] = a12;
    prodimp->arg_names[13] = a13;
    prodimp->arg_names[14] = a14;
    prodimp->arg_names[15] = a15;
    prodimp->arg_names[16] = a16;
    prodimp->arg_names[17] = a17;
    prodimp->arg_names[18] = a18;
    prodimp->arg_names[19] = a19;
    prod_map.insert(Entry(result, prodimp));
  }


  SymbolSetMap buildFollow() const {

    using parse_detail::addTo;

    SymbolSetMap first;
    SymbolSetMap follow;

    // find terminals and add to first set

    for (typename ProdMap::const_iterator i = prod_map.begin();
	 i != prod_map.end();
	 ++i) {

      const Ptr<Production<Context> >& prod = (*i).second;

      const Symbol* b = prod->begin();
      const Symbol* e = prod->end();

      for (Symbol const * j = b;
	   j != e;
	   ++j) {

	Symbol symbol = *j;
	if (isTerminal(symbol)) {
	  first[symbol].insert(symbol);
	}
      }
    }


    bool changed;
    do {

      changed = false;
      for (typename ProdMap::const_iterator i = prod_map.begin();
	 i != prod_map.end();
	 ++i) {

	const Ptr<Production<Context> >& prod = (*i).second;
	assert(prod->getArity()>0);

	Symbol const * j = prod->begin();

	do {

	  Symbol symbol = *j;
	  SymbolSet& first_set = first[symbol];

	  if (j == prod->begin()) {
	    SymbolSet& prod_first_set = first[prod->returnName()];
	    changed = changed || addTo(prod_first_set, first_set);
	  }

	  ++j;

	  if (j!=prod->end()) {

	    Symbol next_symbol = *j;
	    const SymbolSet& next_first_set = first[next_symbol];
	    SymbolSet& follow_set = follow[symbol];
	    changed = changed || addTo(follow_set, next_first_set);

	  } else {

	    // symbol is last symbol

	    SymbolSet& last_follow_set = follow[symbol];
	    SymbolSet& prod_follow_set = follow[prod->returnName()];
	    changed = changed || addTo(last_follow_set, prod_follow_set);
	  }
	} while (j!=prod->end());
      }
    } while (changed);

    return follow;
  }

  ItemSet start() const {

    ItemSet result;

    Range range = prod_map.equal_range(start_prime);

    for (typename ProdMap::const_iterator i = range.first;
	 i != range.second;
	 ++i) {

      result.insert(Item(0, (*i).second));
    }

    return closure(result);
  }


  ItemSet closure(const ItemSet& item_set) const {

    ItemSet result = item_set;
    bool changed;

    do {

      changed = false;

      SymbolSet symbols;

      for (typename ItemSet::const_iterator i = result.begin();
	   i != result.end();
	   ++i) {

	symbols.insert((*i).pending());
      }

      for (SymbolSet::const_iterator j = symbols.begin();
	   j != symbols.end();
	   ++j) {

	Symbol symbol = *j;

	Range range = prod_map.equal_range(symbol);

	typedef typename ProdMap::const_iterator CI;
	for (CI k = range.first;
	     k != range.second;
	     ++k) {

	  changed = result.insert(Item(0,(*k).second)).second || changed;
	}
      }
    } while (changed);

    return result;
  }


  ItemSet goTo(const ItemSet& item_set, Symbol x) const {

    ItemSet result;

    typedef typename ItemSet::const_iterator CI;
    for (CI i = item_set.begin();
	 i != item_set.end();
	 ++i) {

      const Item& item = *i;
      Symbol s = item.pending();

      if (!null(s) && s==x) {
	result.insert(item.advance());
      }
    }
    return closure(result);
  }

  inline bool isTerminal(Symbol s) const {
    // a symbol with no corresponding production is assumed to be a
    // terminal
    Range range = prod_map.equal_range(s);
    return (range.first == range.second);
  }

  Symbol start_symbol;
  Symbol end_symbol;
  Symbol start_prime;
};


} // namespace sigma

#endif
