#include "Cell.hpp"

#ifdef SIGMA_SHOW_COLLECT
#include <iostream>
#endif

#include <assert.h>

#include <vector>
#include <map>

namespace sigma {

Cell* Cell::cells = 0;

bool operator<(const Cell& a, const Cell& b) {
  assert(a.package);
  return a.package->before(b.package);
}

bool operator==(const Cell& a, const Cell& b) {
  return !(a<b) && !(b<a);
}


namespace cell_detail {


bool collecting = false;
Packages* packages = 0;

#ifdef SIGMA_AUTO_COLLECT
bool auto_collect = true;
#else
bool auto_collect = false;
#endif

unsigned int trigger_count = 128;
unsigned int survived = 0;
unsigned int collected = 0;


typedef std::set<Cell*> CellSet;
typedef std::map<Package*,CellSet> CellMap;

void mark(Cell* cell, const CellMap& cell_map) {

  Package* package = cell->package;

  if (package && (package->mark != Package::current_mark)) {

    package->mark = Package::current_mark;

    CellMap::const_iterator i = cell_map.find(package);
    if (i!=cell_map.end()) {

      const CellSet& cell_set = (*i).second;
      for (CellSet::const_iterator j = cell_set.begin();
	   j != cell_set.end();
	   ++j) {

	Cell* next_cell = *j;

	mark(next_cell, cell_map);
      }
    }
  }
}


void init(bool start) {

  if (start) {
    packages = new Packages();
  } else {
    while(collect());
    delete packages;
  }
}

} // end namespace cell_detail


bool collect() {

  using namespace cell_detail;

  if (collecting) return false;
  collecting = true;

  int start_with = packages->size();

  CellMap cell_map;

  typedef std::vector<Cell*> Roots;
  Roots roots;

  /* 
     Build map from packages to the set of contained cells. Also
     build the list of roots.
  */

  for (Cell* cell = Cell::cells; cell != 0; cell = cell->next) {

    if (!null(*cell)) {

      bool root = true;
      for (Packages::iterator j = packages->begin();
	   root && (j != packages->end());
	   ++j) {

	Package* package = *j;

	if (package && package->contains(cell)) {
	  cell_map[package].insert(cell);
	  root = false;
	}
      }

      if (root) {
	roots.push_back(cell);
      }
    }
  }


  Package::current_mark = !Package::current_mark;
  for (Roots::iterator i = roots.begin();
       i != roots.end();
       ++i) {

    mark(*i,cell_map);
  }


  Packages::iterator k = packages->begin();
  while (k != packages->end()) {

    Package* package = *k;

    if (package && (package->mark != Package::current_mark)) {
      delete package;
      Packages::iterator tmp = k;
      ++k;
      packages->erase(tmp);
    } else {
      ++k;
    }
  }


  survived = packages->size();

  collected = start_with-survived;

#ifdef SIGMA_SHOW_COLLECT
  std::cout << "[Collected " 
	    << collected
	    << " out of " 
	    << start_with 
	    << "]" << std::endl;
#endif

  collecting = false;

  return (collected!=0);
}

} // end namespace sigma


