#include "Type.hpp"

#include <assert.h>
#include <iostream>
#include <map>

namespace sigma {


std::ostream& operator<<(std::ostream& out, const Type& type) {

  return out << type.name();
}

namespace {

typedef std::map<Type,std::string> TypeNameMap;
TypeNameMap* type_name_map = 0;

typedef std::map<std::string,Type> TypeMap;
TypeMap* type_map = 0;


} // end namespace

void initTypeNameMap(bool init) {

  if (init) {

    assert(!type_name_map);
    type_name_map = new TypeNameMap();

    assert(!type_map);
    type_map = new TypeMap();

  } else {

    assert(type_name_map);
    delete type_name_map;
    type_name_map = 0;

    assert(type_map);
    delete type_map;
    type_map = 0;
  }
}


void Type::setName(Type type, const std::string& name) {

  assert(type_name_map);
  assert(type_map);

  type_name_map->insert(std::pair<Type,std::string>(type,name));
  type_map->insert(std::pair<std::string,Type>(name,type));
}

Type Type::getTypeByName(const std::string& name) {

  assert(type_map);
  TypeMap::const_iterator it = type_map->find(name);
  if (it==type_map->end()) throw NotFound(name);
  return (*it).second;
}


std::string Type::name() const {

  TypeNameMap::const_iterator it = type_name_map->find(*this);
  if (it != type_name_map->end()) return (*it).second;
  return type_ptr->name();
}



} // end namespace sigma
