#ifndef sigma_ptr_Type_INCLUDED
#define sigma_ptr_Type_INCLUDED

/*
  The Type class associates the appropriate comparison operators with
  type_info pointers. (A reasonable alternative is just to use the raw
  type_info pointers in combination with a type info comparison
  function object.) It supports the ability to specify an alternative
  to the name returned by the type_info structure.
*/

#include <iosfwd>
#include <typeinfo>
#include <string>
#include <init/Init.hpp>


namespace sigma {

struct TypeInfoCompare {
  bool operator ()(const std::type_info* a, const std::type_info* b) const {
    return a->before(*b);
  }
};

struct Type {

  Type() : type_ptr(&typeid(void)) {}
  Type(const std::type_info& t) : type_ptr(&t) {}
  Type(const Type& that) : type_ptr(that.type_ptr) {}

  std::string name() const;

  static void setName(Type, const std::string&);
  static Type getTypeByName(const std::string&);

  // thrown by getTypeByName
  struct NotFound {
    NotFound(const std::string& n) : name(n) {}
    std::string name;
  };

  std::type_info const * type_ptr;
};


inline bool operator==(const Type& a, const Type& b) {
  return (*a.type_ptr) == (*b.type_ptr);
}

inline bool operator!=(const Type& a, const Type& b) {
  return (*a.type_ptr) != (*b.type_ptr);
}

inline bool operator<(const Type& a, const Type& b) {
  return a.type_ptr->before(*b.type_ptr);
}

inline bool operator>(const Type& a, const Type& b) {
  return b.type_ptr->before(*a.type_ptr);
}

inline bool operator<=(const Type& a, const Type& b) {
  return (a<b) || (a==b);
}

inline bool operator>=(const Type& a, const Type& b) {
  return (a>b) || (a==b);
}

std::ostream& operator<<(std::ostream&, const Type&);

template <typename T>
struct DeclareTypeName {
  DeclareTypeName(const char* name) {Type::setName(typeid(T),name);}  
};

void initTypeNameMap(bool);
namespace {Init type_name_map_init(&initTypeNameMap);}

} // end namespace sigma

#endif

