#include "Cons.hpp"

#include <sigma/env.hpp>
#include <sigma/print.hpp>

#include <generic/Generic.hpp>
#include <generic/Primitive.hpp>

namespace sigma {

Ptr<Cons> pr(const Ptr<Anything>& h, const Ptr<Anything>& t) {
  return Ptr<Cons>(new Cons(h, t));
}

Ptr<Anything> hd(const Ptr<Cons>& p) {return p->first;}
Ptr<Anything> tl(const Ptr<Cons>& p) {return p->second;}

Ptr<bool> nil(const Ptr<Anything>& p) {
  return new bool(p.type()==Type(typeid(Nil)));
}


bool isProperList(const Ptr<Cons>& p) {
  if (null(p)) return false;
  Ptr<Anything> t = tl(p);
  if (t.type()==Type(typeid(Nil))) return true;
  return isProperList(t);
}

/*
  The recPrintCons assumes that the list is a proper list and that the
  '[' has already been printed.
*/

Ptr<OutPort> recPrintCons(const Ptr<OutPort>& out, const Ptr<Cons>& p) {

  static Generic print = *getGeneric("print");
  print(out, hd(p));
  Ptr<Anything> t = tl(p);
  if (t.type()==Type(typeid(Nil))) return out << "]";

  out << ", ";
  return recPrintCons(out, t);
}


Ptr<OutPort> printCons(const Ptr<OutPort>& out, const Ptr<Cons>& p) {

  if (isProperList(p)) {
    out << "[";
    return recPrintCons(out, p);
  }

  return out << "pr(" << hd(p) << ", " << tl(p) << ")";
}

Ptr<OutPort> printNil(const Ptr<OutPort>& out, const Ptr<Nil>&) {
  return out << "[]";
}

void initCons(bool init) {

  if (init) {

    {
      Ptr<Generic> generic = getGeneric("pr");
      generic->bind(primitive(&pr));
    }

    {
      Ptr<Generic> generic = getGeneric("hd");
      generic->bind(primitive(&hd));
    }

    {
      Ptr<Generic> generic = getGeneric("tl");
      generic->bind(primitive(&tl));
    }

    {
      Ptr<Generic> generic = getGeneric("nil");
      generic->bind(primitive(&nil));
    }

    {
      Ptr<Generic> generic = getGeneric("print");
      generic->bind(primitive(&printCons));
      generic->bind(primitive(&printNil));

    }
  }
}

} // end namespace sigma
