#ifndef sigma_Lambda_INCLUDED
#define sigma_Lambda_INCLUDED

#include <vector>

#include <symbol/Symbol.hpp>
#include <generic/Method.hpp>
#include <sigma/Formal.hpp>

namespace sigma {

struct Lambda : Method {

  Lambda(const Ptr<Formals> f, const Ptr<Anything>& b) 
    : formals(f), body(b), resolving(false), replacing(false) {}

  Lambda(const Ptr<Formals> f, const Ptr<Anything>& b, Symbol s) 
    : formals(f), body(b), recurse(s), resolving(false) {}

  virtual unsigned int arity() const;
  virtual Type returnType() const; // always Anything
  virtual Type argType(unsigned int) const;

  virtual Ptr<Anything> apply(const Args&) const;

  Ptr<Formals> formals;
  Ptr<Anything> body;

  Symbol recurse;

  bool resolving;
  bool replacing;
  bool evaling;
};


namespace {
DeclareTypeName<Lambda> lambda_type_name("lambda");
DeclareSubtypeConverter<Lambda,Method> lambda_method_type_converter;
}

} // end namespace sigma

#endif
