#ifndef sigma_arith_INCLUDED
#define sigma_arith_INCLUDED

#include <init/Init.hpp>
#include <ptr/Type.hpp>
#include <number/Rational.hpp>
#include <number/Integer.hpp>
#include <number/Natural.hpp>
#include <generic/Generic.hpp>

/*
  Supported forms:

  double + double => double
  Natural + Natural => Natural
  Integer + Integer => Integer
  Rational + Rational => Rational

  double - double => double
  Integer - Integer => Integer
  Rational - Rational => Rational

  double * double => double
  Natural * Natural => Natural
  Integer * Integer => Integer
  Rational * Rational => Rational

  double / double => double
  Rational / Rational => Rational

  - double => double
  - Integer => Integer
  - Rational => Rational

  also: > < =

*/

namespace sigma {

namespace {

DeclareTypeName<bool> bool_type_name("boolean");
DeclareTypeName<Natural<> > natural_type_name("natural");
DeclareTypeName<Integer<> > integer_type_name("integer");
DeclareTypeName<Rational<> > rational_type_name("rational");
DeclareTypeName<double> double_type_name("real");

DeclareCastConverter<Natural<>, Integer<> > natural_integer_converter;
DeclareCastConverter<Integer<>, Rational<> > integer_rational_converter;
DeclareCastConverter<Rational<>, double> rational_double_converter;

// type cast short cuts
DeclareCastConverter<Natural<>, double> natural_double_converter;
DeclareCastConverter<Integer<>, double> integer_double_converter;

} // end namespace

void initArith(bool);
namespace {Init arith_init(&initArith);}


} // end namespace sigma

#endif
