#ifndef sigma_env_INCLUDED
#define sigma_env_INCLUDED

#include <ptr/Ptr.hpp>
#include <ptr/Ptr.hpp>
#include <env/Env.hpp>
#include <symbol/Symbol.hpp>
#include <generic/Generic.hpp>

namespace sigma {

typedef Env<Symbol,Ptr<Anything> > SigmaEnv;

// support for first-class environments

struct ThisEnv {

  ThisEnv() {}
  ThisEnv(const Ptr<SigmaEnv>& e) : env(e) {}

  Ptr<SigmaEnv> env;
};


// Can't be initialized automatically because of initialization order
// issues. Note that the environment won't automatically be cleaned up
// on exit. It must be done explicitly (assuming the application
// cares). (I hate this, but I see no easy way around it. At least
// it's the only exception.)

Ptr<SigmaEnv>& theEnv();

// Gets a generic out of the environment. Adds one to the environment
// if it hasn't been added yet (or if the symbol is bound to something
// else). (Note that Generic has object semantics.)

Ptr<Generic> getGeneric(Symbol);

// An empty environment can't really be a null pointer or generic
// dispatching on its type doesn't work. So an empty environment
// actually contains a null symbol associated with a null pointer.

inline Ptr<SigmaEnv> make_sigma_env() {
  return new SigmaEnv(Symbol(),Ptr<Nothing>());
}

namespace {
DeclareTypeName<SigmaEnv> sigmaenv_type_name("env");
DeclareTypeName<ThisEnv> thisenv_type_name("this");
}

} // end namespace sigma

#endif
