#ifndef sigma_sigma_lex_INCLUDED
#define sigma_sigma_lex_INCLUDED

#include <assert.h>
#include <string>
#include <lex/Dfa.hpp>
#include <parse/ProdVal.hpp>

#include <sigma/Context.hpp>


/*

  term_op is separated into pos_op and neg_op so that a unary minus
  can be defined

  let_lit
  fun_lit
  fn_lit
  if_lit
  be_lit
  then_lit
  else_lit
  this_lit
  natural
  real
  ident
  string
  )
  (
  [
  ]
  and_op
  or_op
  comp_op
  factor_op
  +
  -
  ;
  :
  ,
  {
  }
  '
  end_of_input

*/


namespace sigma {

typedef Ptr<dfa_detail::Node<char,ProdVal,Context> > RegExp;

RegExp token(char);
RegExp range(char from, char to);
RegExp literal(const char*);
RegExp oneof(const char*);
 RegExp rule(RegExp regexp, Action<char,ProdVal,Context> action);

RegExp buildRegExp();

// Translates escape sequences to their corresponding characters.
std::string fromEscSeq(const std::string&); 

// Translates unprintable characters to escape sequences.
std::string toEscSeq(const std::string&);   

namespace {
DeclareTypeName<std::string> string_type_name("string");
} // namespace


} // namespace sigma

#endif
