#ifndef sigma_parse_INCLUDED
#define sigma_parse_INCLUDED

#include <ptr/Cell.hpp>
#include <parse/Productions.hpp>
#include <sigma/Context.hpp>

/*
  statement -> binding			// identity
  binding -> 'let' ident 'be' exp	// bind

  statement -> funbind			// identity
  funbind -> 'let' ident proto 'be' exp	// funbind

  statement -> exp			// identity

  exp -> if_exp				// identity
  if_exp -> 'if' exp 'then' exp 'else' 	// if
  if_exp -> bool_exp			// identity

  bool_exp -> bool_exp "&" comp_exp	// andop
  bool_exp -> bool_exp "|" comp_exp	// orop
  bool_exp -> comp_exp			// identity

  // since these return bool, don't bother allowing chaining

  comp_exp -> term_exp comp_op term_exp	// binop
  comp_exp -> term_exp			// identity

  term_exp -> term_exp '-' factor_exp	// binop
  term_exp -> term_exp '+' factor_exp	// binop
  term_exp -> factor_exp		// identity

  factor_exp -> factor_exp factor_op factor	// binop
  factor_exp -> factor				// identity

  factor -> string			// identity
  factor -> natural			// identity
  factor -> real			// identity
  factor -> ident			// identity
  factor -> this			// this_op
  factor -> '-' natural			// unop
  factor -> '-' ident			// unop
  factor -> '(' exp ')'			// paren
  factor -> '[' items ']'		// list
  factor -> '[' ']'			// nil
  factor -> block			// identity

  items -> exp ',' items		// items
  items -> exp				// item

  factor -> factor '(' args ')'		// apply
  factor -> factor '(' ')'		// apply_0
  args	-> args ',' exp			// args
  args 	-> exp				// arg

  factor -> 'fn' proto block		// fn
  proto -> '(' formals ')'		// proto
  proto -> '(' ')'			// proto_0

  factor -> ' exp ';			// unevaled exp

  formals -> formals ',' formal		// formals
  formals -> formal			// formals_1
  formal -> ident ':' ident		// typed_formal
  formal -> ident			// formal


  block -> '{' statements '}'		// block
  statements -> statement ',' statements  // statements
  statements -> statement		// statement

*/

namespace sigma {

Productions<Context> buildProductions();

} // end namespace sigma

#endif
