#include "print.hpp"

#include <iostream>

#include <sigma/lex.hpp>
#include <sigma/env.hpp>

#include <ptr/Ptr.hpp>
#include <number/Natural.hpp>
#include <number/Integer.hpp>
#include <number/Rational.hpp>
#include <generic/Primitive.hpp>

/*
  This <sigma/arith.h> needs to be included before the print generic
  function is initialized to guarantee that the DeclareCastConverters
  for the number types establishes the appropriate 'isa' relationships
  among the number types.

  It may be more appropriate to move the initialization of the number
  print routines directly into arith.cpp.
*/

#include <sigma/arith.hpp>


namespace sigma {

Ptr<OutPort> operator<<(const Ptr<OutPort>& out, const Ptr<Anything>& item) {

  static Generic print = *getGeneric("print");

  print(out, item);
  return out;
}

OutPort::OutPort() : out(&std::cout) {}

namespace {

void initPrintGeneric(bool);
Init init_print_generic(&initPrintGeneric);

Ptr<OutPort> std_outport;


Ptr<OutPort> printAnything(const Ptr<OutPort>& out, const Ptr<Anything>& p) {

  return out << "<" << p.type().name() << ">";
}


Ptr<OutPort> printBool(const Ptr<OutPort>& out, const Ptr<bool>& item) {
  if (*item) {
    out << "true";
  } else {
    out << "false";
  }
  return out;
}


Ptr<OutPort> printString(const Ptr<OutPort>& out, 
			 const Ptr<std::string>& item) {

  return out << "\"" << toEscSeq(*item) << "\"";
}


Ptr<Nothing> printAnythingStdOutPort(const Ptr<Anything>& item) {

  if (null(std_outport)) {
    std_outport = new OutPort();
  }

  static Generic print = *getGeneric("print");

  print(std_outport, item);

  return Ptr<Nothing>();
}


/*
  This doesn't work with circular lists.
*/

template <typename T>
Ptr<OutPort> printT(const Ptr<OutPort>& out, const Ptr<T>& item) {
  return out << *item;
}


void initPrintGeneric(bool init) {

  if (init) {

    Ptr<Generic> print = getGeneric("print");

    print->bind(primitive(&printAnything));
    print->bind(primitive(&printBool));
    print->bind(primitive(&printString));
    print->bind(primitive(&printT<double>));
    print->bind(primitive(&printT<Rational<> >));
    print->bind(primitive(&printT<Integer<> >));
    print->bind(primitive(&printT<Natural<> >));
    print->bind(primitive(&printAnythingStdOutPort));
  }
}

} // end namespace


} // end namespace sigma
