#include "replace.hpp"

#include <init/Init.hpp>
#include <ptr/Ptr.hpp>
#include <sigma/env.hpp>

#include <symbol/Symbol.hpp>
#include <generic/Primitive.hpp>
#include <env/Env.hpp>
#include <sigma/env.hpp>
#include <sigma/Block.hpp>
#include <sigma/Cons.hpp>
#include <sigma/Bind.hpp>
#include <sigma/Lambda.hpp>
#include <sigma/Apply.hpp>
#include <sigma/If.hpp>


namespace sigma {

Generic replace("replace");

namespace {

// The eval, replace, and resolve generics have to be initialized
// after the components of the parse tree. That's why the init
// function isn't in the header. (Generics only have object semantics
// after they've been initialized.)


void initReplaceGeneric(bool);
Init init_replace_generic(&initReplaceGeneric);

Ptr<Anything> replaceAnything(const Ptr<RecLambda>&,
			      const Ptr<Anything>& anything) {

  return anything;
}


Ptr<Anything> replaceSymbol(const Ptr<RecLambda>& rec_lambda_ptr, 
			    const Ptr<Symbol>& symbol_ptr) {

  if (rec_lambda_ptr->name == *symbol_ptr)
    return rec_lambda_ptr->lambda;

  return symbol_ptr;
}


Ptr<Anything> replaceArgs(const Ptr<RecLambda>& rec_lambda,
			  const Ptr<Args>& args_ptr) {

  Args& args = deref<Args>(args_ptr);
  for (Args::iterator i = args.begin();
       i != args.end();
       ++i) {

    *i = replace(rec_lambda, *i);
  }

  return args_ptr;
}


Ptr<Anything> replaceApply(const Ptr<RecLambda>& rec_lambda,
			   const Ptr<Apply>& apply_ptr) {

  Apply& apply = deref<Apply>(apply_ptr);
  apply.func = sigma::replace(rec_lambda, apply.func);
  apply.args = sigma::replace(rec_lambda, apply.args);

  return apply_ptr;
}
  


Ptr<Anything> replaceLambda(const Ptr<RecLambda>& rec_lambda_ptr,
			    const Ptr<Lambda>& lambda_ptr) {

  Lambda& lambda = deref<Lambda>(lambda_ptr);
  RecLambda& rec_lambda = deref<RecLambda>(rec_lambda_ptr);

  if (lambda.replacing) return lambda_ptr;

  // Don't recursively replace symbols if shadowed by a lambda free
  // variable.

  const Formals& formals = *lambda.formals;
  Symbol name = rec_lambda.name;

  for (Formals::const_iterator i = formals.begin();
       i != formals.end();
       ++i) {

    if ((*i).symbol==name) return lambda_ptr;
  }


  lambda.replacing = true;

  lambda.body = replace(rec_lambda_ptr, lambda.body);

  lambda.replacing = false;

  return lambda_ptr;
}


Ptr<Anything> replaceBlock(const Ptr<RecLambda>& rec_lambda_ptr,
			   const Ptr<Block>& block_ptr) {

  Block& block = deref<Block>(block_ptr);
  block.statement = replace(rec_lambda_ptr, block.statement);
  block.rest = replace(rec_lambda_ptr, block.rest);

  return block_ptr;
}


Ptr<Anything> replaceCons(const Ptr<RecLambda>& rec_lambda_ptr,
			  const Ptr<Cons>& cons_ptr) {

  Cons& cons = deref<Cons>(cons_ptr);

  cons.first = replace(rec_lambda_ptr, cons.first);
  cons.second = replace(rec_lambda_ptr, cons.second);

  return cons_ptr;
}


Ptr<Anything> replaceIf(const Ptr<RecLambda>& rec_lambda_ptr,
			const Ptr<If>& if_ptr) {

  If& the_if = deref<If>(if_ptr);
  the_if.exp0 = replace(rec_lambda_ptr, the_if.exp0);
  the_if.exp1 = replace(rec_lambda_ptr, the_if.exp1);
  the_if.exp2 = replace(rec_lambda_ptr, the_if.exp2);

  return if_ptr;
}

void initReplaceGeneric(bool init) {

  if (init) {
    replace.bind(primitive(&replaceAnything));
    replace.bind(primitive(&replaceSymbol));
    replace.bind(primitive(&replaceArgs));
    replace.bind(primitive(&replaceApply));
    replace.bind(primitive(&replaceLambda));
    replace.bind(primitive(&replaceBlock));
    replace.bind(primitive(&replaceCons));
    replace.bind(primitive(&replaceIf));
  }
}

} // end namespace

} // end names
