#ifndef sigma_resolve_INCLUDED
#define sigma_resolve_INCLUDED

#include <generic/Generic.hpp>

/*
  Resolve has two jobs. The first is to replace all bound identifiers
  with their values. The second is to transform bindings within blocks
  into apply/eval constructs and note free variables as it descends
  into the AST. The result is an AST without any explicit binding
  constructs, and no bound variables--only free ones. This has the
  side-effect of checking for the use of undefined variables.

  Supported forms:

  (SigmaEnv, Anything) -> Anything
  (SigmaEnv, Symbol) -> Anything
  (SigmaEnv, Args) -> Anything
  (SigmaEnv, Apply) -> Anything
  (SigmaEnv, Lambda) -> Anything
  (SigmaEnv, Block) -> Anything
  (SigmaEnv, Cons) -> Anything
  (SigmaEnv, If) -> Anything
  (SigmaEnv, Bind) -> Anything
  (SigmaEnv, ThisEnv) -> Anything

  In theory, a bind can only happen in a block or at the top level
  (because it's not an expression), and when it happens in a block
  it's transformed by the resolveBlock method. Since resolve is only
  called as a result of evaluating a(n unapplied) lambda, resolve
  should never encounter a Bind construct.

*/


namespace sigma {

extern Generic resolve;

struct Free {}; // denote free variable (also used by bind)

} // end namespace sigma


#endif
