#include "Symbol.hpp"

#include <assert.h>

#include <iostream>

namespace sigma {

Symbol gensym(const char* name) {

  assert(Symbol::table);
  int tag = Symbol::table->names.size();
  Symbol::table->names.push_back(name);
  return Symbol(tag);
}


Symbol gensym(const std::string& name) {

  assert(Symbol::table);
  int tag = Symbol::table->names.size();
  Symbol::table->names.push_back(name);
  return Symbol(tag);
}


std::ostream& operator<<(std::ostream& out, const Symbol& symbol) {
  return out << symbol.name();
}

Symbol::Table* Symbol::table = 0;

void Symbol::init(bool starting) {
  if (starting) {
    assert (table==0);
    table = new Table();
    std::string empty_str;
    std::pair<Table::Tags::iterator, bool> i 
      = table->tags.insert(std::pair<std::string,int>(empty_str,0));
    assert(i.second);
    table->names.push_back(empty_str);
  } else {
    delete table;
  }
}

int Symbol::getTag(const std::string& name) {

  assert(table);

  Table::Tags::iterator i = table->tags.lower_bound(name);

  if ((i == table->tags.end()) || ((*i).first != name)) {
    int tag = table->names.size();
    table->tags.insert(i, std::pair<const std::string,int>(name,tag));
    table->names.push_back(name);
    return tag;
  } else {
    return (*i).second;
  }
}


const std::string& Symbol::name() const {
  
  assert(table);
  assert(tag <= (table->names.size()));
  return table->names[tag];
}


} // end namespace sigma
