#ifndef sigma_Symbol_INCLUDED
#define sigma_Symbol_INCLUDED

#include <iosfwd>
#include <map>
#include <vector>
#include <string>

#if 0
#include <utility>
using namespace std::rel_ops;
#endif


#include <init/Init.hpp>

namespace sigma {

struct Symbol {

  Symbol() : tag(0) {}
  Symbol(const Symbol& that) : tag(that.tag) {}
  Symbol(const std::string& name) : tag(getTag(name)) {}
  Symbol(const char* name) : tag(getTag(name)) {}

  friend inline bool operator==(Symbol a, Symbol b) {
    return a.tag==b.tag;
  }
  friend inline bool operator<(Symbol a, Symbol b) {
    return a.tag<b.tag;
  }
  friend inline bool null(Symbol s) {
    return s.tag==0;
  }

  const std::string& name() const;

  static void init(bool);

  struct Table {

    typedef std::map<std::string, int> Tags;
    Tags tags;

    typedef std::vector<std::string> Names;
    Names names;
  };

 private:
    
  unsigned int tag;

  int getTag(const std::string& name);

  static Table* table;

 private:

/*
  The gensym function generates a symbol that is guaranteed to be
  unique even if its name isn't.
*/

  friend Symbol gensym(const std::string&);
  friend Symbol gensym(const char*);

  Symbol(int t) : tag(t) {}
};


std::ostream& operator<<(std::ostream&, const Symbol&);

namespace {
  Init symbol_init(&Symbol::init);
}

} // end namespace sigma

#endif
