#include <iostream>

#include <symbol/Symbol.hpp>
#include <dag/Dag.hpp>

using namespace sigma;

struct Edge {

  Edge() {}
  Edge(const char* f, const char* t) : from(f), to(t) {}

  Symbol from;
  Symbol to;

};

typedef Dag<Symbol,Edge> MyDag;


std::ostream& operator<<(std::ostream& out, const MyDag::Path& path) {

  for (MyDag::Path::const_iterator i = path.begin();
       i != path.end();
       ++i) {

    const Edge& edge(*i);
    out << "from: " << edge.from << "\t to: " << edge.to << std::endl;
  }
  return out;
}


int main() {

  try {

    MyDag dag;

    dag.insert("natural","integer",Edge("natural","integer"));
    dag.insert("integer","rational",Edge("integer","rational"));
    dag.insert("rational","real",Edge("rational","real"));
    dag.insert("real","complex",Edge("real","complex"));

    MyDag::Path path = dag.getPath("integer","complex");
    std::cout << "path from integer to complex: " 
	      << std::endl << path << std::endl;

    dag.insert("integer","real",Edge("integer","real"));

    MyDag::Path path2 = dag.getPath("integer","complex");
    std::cout << "path from integer to complex: " 
	      << std::endl << path2 << std::endl;

    dag.insert("complex","integer",Edge("complex","integer"));

  }

  catch (const MyDag::Circular& error) {

    std::cout << "Circular(" << error.from 
	      << "," << error.to << ")" << std::endl;
  }

  return 0;
}
