
#include <string>
#include <iostream>

#include <ptr/Type.hpp>

#include <generic/Method.hpp>
#include <generic/Generic.hpp>
#include <generic/Primitive.hpp>

using namespace sigma;
using namespace std;


Ptr<float> processingFloats(const Ptr<float>& a, 
			    const Ptr<float>& b) {

  cout << "processing floats(" 
       << *a.get() << "," 
       << *b.get() << ")" << endl;
  return new float(*a.get() + *b.get());
}


Ptr<double> processingDoubles(const Ptr<double>& a,
			      const Ptr<double>& b) {

  cout << "processing doubles(" 
       << *a.get() << "," 
       << *b.get() << ")" << endl;
  return new double(*a.get() + *b.get());
}

Ptr<int> processingInts(const Ptr<int>& a, 
			const Ptr<int>& b) {

  cout << "processing ints(" 
       << *a.get() << "," 
       << *b.get() << ")" << endl;
  return new int(*a.get() + *b.get());
}

Ptr<Anything> processingAny(const Ptr<Anything>& a, 
			    const Ptr<Anything>& b) {

  cout << "processing any(" 
       << a.type().name() << "," 
       << b.type().name() << ")" << endl;
  return Cell();
}

DeclareCastConverter<int,float> int_float_converter;
DeclareCastConverter<float,double> float_double_converter;
DeclareSubtypeConverter<Primitive<Ptr<float>(*)(const Ptr<float>&,
						const Ptr<float>&)>,Method> pf;
						  


int main() {

  Generic generic;

  Type f_type = typeid(Method);
  Type p_type = typeid(Primitive<Ptr<float>(*)(const Ptr<float>&,
					       const Ptr<float>&)>);

  std::cout << "isa(" << f_type.name() << ", " << p_type.name()
	    << ") :"
	    << isa(p_type, f_type)
	    << endl;

  Ptr<Method> pi = primitive(&processingInts);
  generic.bind(pi);

  Ptr<Method> pa = primitive(&processingAny);
  generic.bind(pa);


  Ptr<Method> pf = primitive(&processingFloats);
  generic.bind(pf);

  Ptr<Method> pd = primitive(&processingDoubles);
  generic.bind(pd);


  {
    Args args;
    args.push_back(new int(2));
    args.push_back(new int(3));
    generic.dispatch(args)->apply(args);
  }

  {
    Args args;
    args.push_back(new int(2));
    args.push_back(new float(3.4f));
    generic.dispatch(args)->apply(args);
  }
  {
    Args args;
    args.push_back(new float(2.3f));
    args.push_back(new int(3));
    generic.dispatch(args)->apply(args);
  }
  {
    Args args;
    args.push_back(new float(2.3f));
    args.push_back(new double(3.4));
    generic.dispatch(args)->apply(args);
  }
  {
    Args args;
    args.push_back(new std::string("hi there"));
    args.push_back(new float(2.3f));
    generic.dispatch(args)->apply(args);
  }


  {
    double* a_raw = new double(6);
    double* b_raw = new double(7);
    Ptr<double> a(a_raw);
    Ptr<double> b(b_raw);
    std::cout << "inline dispatch(" 
	      << a_raw << "," << b_raw << ")" << std::endl;
    
    generic(a,b);
  }

}
